#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <errno.h>

#include <sys/time.h>
#include <time.h>

#define osErrorFatal(userMsg) osErrorFatalImpl((userMsg), __FILE__, __func__, __LINE__)
#define osAssert(expr, userMsg) \
    do { \
        if (!(expr)) \
            osErrorFatal(userMsg); \
    } while(0)
    
static const char *os_Usage = "";
    
void osErrorFatalImpl(const char *userMsg, const char *fileName, 
                      const char *functionName, const int lineNum);

int main(int argc, char** argv) {
	
	/* ucitavamo vreme u sekundama */
	time_t now = time(NULL);
	/* ucitavamo vreme u milisekundama */
	struct timeval nowInUsec;
	osAssert(gettimeofday(&nowInUsec, NULL) != -1, "gettimeofday failed");
	
	/* proveravamo da li je vreme iz jednog i iz drugog poziva jednako */
	if (now == nowInUsec.tv_sec) {
		printf("Vreme je jednako\n");
	}
	else {
		printf("Vreme nije jednako\n");
	}
	/* stampamo razliku vremena */
	printf("Izgubili smo: %jd\n", (intmax_t)nowInUsec.tv_usec);
	
	/* prebacujemo vreme iz sekundi u format citljiv korisniku */
	struct tm* timeStruct = localtime(&now);
	osAssert(timeStruct != NULL, "localtime failed");
	/* uvecamo dan u mesecu za jedan
	 * 
	 * BITNO:
	 * Voditi racuna da se od programera ocekuje da vodi racuna
	 * o konzistentnosti vremena u strukturi tm
	 * Programer ne sme da dozvoli da postoji npr. 31.2.2017
	 */ 
	timeStruct->tm_mday++;
	/* prebacivanje vremena iz citljivog formata nazad u broj sekundi */
	time_t newTime = mktime(timeStruct);
	
	/* prikaz novodobijenog datuma */
	printf("%s", ctime(&newTime));
	/* ukupan broj sekundi u jednom danu */
	printf("Razlika u sekundama: %jd\n", (intmax_t)(newTime - now));
	
	exit(EXIT_SUCCESS);
}

void osErrorFatalImpl(const char *userMsg, const char *fileName, 
                      const char *functionName, const int lineNum) {
    perror(userMsg);
    fprintf(stderr, "File: '%s'\nFunction: '%s'\nLine: '%d'\n", fileName, functionName, lineNum);
    exit(EXIT_FAILURE);
}
